import dpath.util as dpath
import sys
import os
import click
import uuid
from . import writer
from requests.compat import urlsplit
from functools import update_wrapper
from importlib import import_module


def is_uuid(str):
    try:
        uuid.UUID(str)
        return True
    except ValueError:
        return False


def lookup_resource_id(list, id_rep, name_path=None, mac=False, **kwargs):
    if hasattr(list, '__call__'):
        list = list().get('data')
    _is_uuid = not mac and is_uuid(id_rep)
    id_rep_lower = id_rep.lower()
    id_rep_len = len(id_rep)
    name_path = name_path or "attributes/name"
    matches = []
    for entry in list:
        entry_id = entry.get('id')
        if _is_uuid:
            if entry_id == id_rep:
                return entry_id
        elif mac:
            try:
                entry_mac = dpath.get(entry, 'meta/mac')
                if entry_mac[-id_rep_len:].lower() == id_rep_lower:
                    matches.append(entry_id.encode('utf8'))
            except KeyError:
                pass

        else:
            short_id = shorten_id(entry_id)
            if short_id == id_rep:
                matches.append(entry_id.encode('utf8'))
            else:
                try:
                    entry_name = dpath.get(entry, name_path)
                    if entry_name[:id_rep_len].lower() == id_rep_lower:
                        matches.append(entry_id.encode('utf8'))
                except KeyError:
                    pass
    if len(matches) == 0:
        raise KeyError('Id: ' + id_rep.encode('utf8') + ' does not exist')
    elif len(matches) > 1:
        short_matches = [shorten_id(id) for id in matches]
        match_list = ' (' + ', '.join(short_matches) + ')'
        raise KeyError('Ambiguous id: ' + id_rep.encode('utf8') + match_list)

    return matches[0]


def shorten_id(str):
    return str.split('-')[0]


def shorten_json_id(json, **kwargs):
    # Ugh, reaching for global state isn't great but very convenient here
    _uuid = kwargs.get('uuid', False)
    try:
        root_context = click.get_current_context().find_root()
        shorten = not root_context.params.get('uuid', False)
    except RuntimeError:
        shorten = not _uuid
    json_id = json.get('id')
    return shorten_id(json_id) if shorten else json_id


def tabulate(result, map, **kwargs):
    if not map or not result:
        return result

    file = kwargs.pop('file', click.utils.get_text_stream('stdout'))
    with writer.for_format(output_format(**kwargs),
                           file, mapping=map, **kwargs) as _writer:
        _writer.write_entries(result)


def map_script_filenames(json):
    files = dpath.get(json, 'meta/scripts')
    return ', '.join(extract_script_filenames(files))


def extract_script_filenames(files):
    return [urlsplit(url).path.split('/')[-1] for url in files]


def output_format(default_format='tabular', **kwargs):
    override_format = kwargs.get('format')
    try:
        root_context = click.get_current_context().find_root()
        click_format = root_context.params.get('format')
    except RuntimeError:
        click_format = None
    return override_format or click_format or default_format


def sort_option(options):
    options = [
        click.option('--reverse', is_flag=True,
                     help='Sort in reverse order'),
        click.option('--sort', type=click.Choice(options),
                     help='How to sort the result')
    ]

    def wrapper(func):
        for option in reversed(options):
            func = option(func)
        return func
    return wrapper


CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help']
)


def cli(version=None, package=None,  commands=None):
    class Loader(click.MultiCommand):
        def list_commands(self, ctx):
            commands.sort()
            return commands

        def get_command(self, ctx, name):
            try:
                command = import_module(package + "." + name)
                return command.cli
            except ImportError as e:
                click.secho(str(e), fg='red')
                return

    def decorator(f):
        @click.option('--uuid', is_flag=True,
                      help="Whether to display long identifiers")
        @click.option('--format',
                      type=click.Choice(['csv', 'json', 'tabular']),
                      default=None,
                      help="The output format (default 'tabular')")
        @click.version_option(version=version)
        @click.command(cls=Loader, context_settings=CONTEXT_SETTINGS)
        @click.pass_context
        def new_func(ctx, *args, **kwargs):
            ctx.invoke(f, ctx, *args, **kwargs)
        return update_wrapper(new_func, f)
    return decorator


def main(cli):
    def decorator():
        args = sys.argv[1:]
        try:
            cli.main(args=args, prog_name=None)
        except Exception as e:
            if os.environ.get("HELIUM_COMMANDER_DEBUG"):
                raise
            click.secho(str(e), fg='red')
            sys.exit(1)
    return decorator



from helium.version import __version__
try:
    # ignore import errors for setup.py to be able to do it's work
    # pre-dependency installations
    from helium.service import Service
except:
    pass



import click
import helium
from . import timeseries as ts
from .sensor import version_option as sensor_version_option
from .sensor import mac_option, sort_option as sensor_sort_option
from .sensor import _tabulate as _tabulate_sensors
from .util import tabulate, lookup_resource_id, shorten_json_id
from .util import sort_option as _sort_option
import dpath.util as dpath

pass_service = click.make_pass_decorator(helium.Service)


def version_option(f):
    return click.option('--versions', type=click.Choice(['none', 'fw']),
                        default='none',
                        help="display version information")(f)


def sort_option(f):
    return _sort_option(['name'])(f)


@click.group()
def cli():
    """Operations on elements.
    """
    pass


def _find_element_id(service, element, **kwargs):
    return lookup_resource_id(service.get_elements, element, **kwargs)


def _tabulate(result, **kwargs):
    def _map_sensor_count(json):
        return len(dpath.get(json, 'relationships/sensor/data'))
    version_map = []
    version_option = kwargs.pop('versions', 'none')
    if version_option == 'fw':
        version_map = [
            ('firmware', 'meta/versions/element'),
        ]
    tabulate(result, [
        ('id', shorten_json_id),
        ('mac', 'meta/mac'),
        ('sensors', _map_sensor_count)
    ] + version_map + [
        ('name', 'attributes/name')
    ], **kwargs)


@cli.command()
@click.argument('element', required=False)
@version_option
@mac_option
@sort_option
@pass_service
def list(service, element, **kwargs):
    """List elements.

    Lists one or all elements in the organization.
    """
    if element:
        element = _find_element_id(service, element, **kwargs)
        elements = [service.get_element(element, include='sensor').get('data')]
    else:
        elements = service.get_elements(include='sensor').get('data')
    _tabulate(elements, **kwargs)


@cli.command()
@click.argument('element')
@click.option('--name',
              help="the new name for the sensor")
@mac_option
@pass_service
def update(service, element, **kwargs):
    """Updates the attributes of an element.

    Updates the attributes of a given ELEMENT.
    """
    element = _find_element_id(service, element, **kwargs)
    data = service.update_element(element, **kwargs).get('data')
    _tabulate([data])


@cli.command()
@click.argument('element')
@sensor_version_option
@sensor_sort_option
@mac_option
@pass_service
def sensor(service, element, **kwargs):
    """Lists sensors for an element.

    Lists the sensors for a given ELEMENT.
    """
    element = _find_element_id(service, element, **kwargs)
    sensors = service.get_element(element, include='sensor').get('included')
    _tabulate_sensors(sensors, **kwargs)


@click.argument('element')
@mac_option
@pass_service
def _get_element_timeseries(service, element, **kwargs):
    """Get timeseries readings for a element.

    Retrieve timeseries data for a given ELEMENT.
    """
    element = _find_element_id(service, element, **kwargs)
    return service.get_element_timeseries(element, **kwargs).get('data')


@click.argument('element')
@mac_option
@pass_service
def _post_element_timeseries(service, element, **kwargs):
    """Post readings to a element.

    Posts timeseries readings for a given ELEMENT.
    """
    element = _find_element_id(service, element, **kwargs)
    return [service.post_element_timeseries(element, **kwargs).get('data')]


@click.argument('element')
@mac_option
@pass_service
def _live_element_timeseries(service, element, **kwargs):
    """Live readings from a element.

    Reports readings from a given ELEMENT as they come in.
    """
    element = _find_element_id(service, element, **kwargs)
    return service.live_element_timeseries(element, **kwargs)


cli.add_command(ts.cli(get=_get_element_timeseries,
                       post=_post_element_timeseries,
                       live=_live_element_timeseries))



import io
import os
from . import __version__
from json import loads as load_json, dumps as dump_json
from requests import Session, HTTPError
from requests.compat import urljoin, urlsplit


class LiveService:
    """Represents a live SSE endpoint.

    Some of the helium endpoints return a Server Sent Event connection
    which returns events as they happen in the system. The most common
    usecase for this are the "live" endpoints.

    Typically you work with this class by calling one of the Service
    endpoints which returns an instance of this and then iterate over
    the result of the `events` method.

    """
    _FIELD_SEPARATOR = ':'

    def __init__(self, source):
        """Initializes a LiveSession.

        :param source: A requests.Response object
        """
        self.source = source

    def _read(self):
        data = ""
        for line in self.source.iter_lines(decode_unicode=True):
            if not line.strip():
                yield data
                data = ""
            data = data + "\n" + line

    def events(self):
        for chunk in self._read():
            event_type = None
            event_data = ""
            for line in chunk.splitlines():
                # Ignore empty lines or comments
                # Comments lines in SSE start with the field separator
                if not line.strip() or line.startswith(self._FIELD_SEPARATOR):
                    continue

                data = line.split(self._FIELD_SEPARATOR, 1)
                field = data[0]
                data = data[1]

                if field == 'event':
                    event_type = data
                elif field == 'data':
                    event_data += data
                else:
                    event_data = data

            if not event_data:
                # Don't report on events with no data
                continue

            if event_data.endswith('\n'):
                event_data = event_data[:-1]

            event_data = load_json(event_data)
            yield (event_type, event_data)

    def close(self):
        self.source.close()


class Service:
    production_base_url = "https://api.helium.com/v1"
    user_agent = "Helium/"+__version__

    def __init__(self, api_key, base_url=None):
        self.api_key = api_key
        self.base_url = base_url if base_url else self.production_base_url
        # ensure we can urljoin paths to the base url
        if not self.base_url.endswith('/'):
            self.base_url += '/'
        self.session = Session()

    def mk_params(self, map, kwargs):
        result = {}
        for kw_key, result_key in map.iteritems():
            value = kwargs.get(kw_key, None)
            if value:
                result[result_key] = value
        return result

    def mk_attributes_body(self, type, id, attributes):
        if attributes is None or not type:
            return None
        result = {
            "data": {
                "attributes": attributes,
                "type": type
            }
        }
        if id is not None:
            result['data']['id'] = id
        return result

    def mk_relationships_body(self, type, ids):
        if ids is None or not type:
            return None
        return {
            "data": [{"id": id, "type": type} for id in ids]
        }

    def mk_datapoint_body(self, **kwargs):
        params = self.mk_params({
            'value': 'value',
            'port': 'port',
            'timestamp': 'timestamp',
        }, kwargs)
        return self.mk_attributes_body("data-point", None, params)

    def mk_timeseries_params(self, **kwargs):
        return self.mk_params({
            "page_id": "page[id]",
            "page_size": "page[size]",
            "port": "filter[port]",
            "start": "filter[start]",
            "end": "filter[end]",
            "agg_size": "agg[size]",
            "agg_type": "agg[type]"
        }, kwargs)

    def mk_include_params(self, **kwargs):
        return self.mk_params({
            "include": "include"
        }, kwargs)

    def do(self, method, path,
           params=None, json=None,
           files=None, stream=False):
        if path is None:
            return None
        url = path if urlsplit(path).scheme else urljoin(self.base_url, path)
        params = params or {}
        headers = {
            'Authorization': self.api_key,
            'User-agent': self.user_agent
        }
        res = self.session.request(method, url,
                                   stream=stream,
                                   params=params,
                                   json=json,
                                   files=files,
                                   headers=headers,
                                   allow_redirects=True)
        if res.ok:
            try:
                return res.json() if not stream else res
            except ValueError:
                return res
        else:
            try:
                # Pull out the first error info block (for now)
                errors = res.json()['errors']
                err = errors[0]
                errmsg = str.format("{} Error: {} for url {}",
                                    err['status'],
                                    err['detail'],
                                    res.url)
                # and use that as the error
                raise HTTPError(errmsg, response=res)
            except ValueError:
                # otherwise raise as a base HTTPError
                res.raise_for_status()

    def get(self, path,  **kwargs):
        return self.do('GET', path, **kwargs)

    def post(self, path, **kwargs):
        return self.do('POST', path, **kwargs)

    def delete(self, path, **kwargs):
        return self.do('DELETE', path, **kwargs)

    def patch(self, path, **kwargs):
        return self.do('PATCH', path, **kwargs)

    def is_production(self):
        return self.base_url == self.production_base_url

    def get_user(self):
        return self.get('user')

    def auth_user(self, user, password):
        body = {
            "email": user,
            "password": password
        }
        return self.post('user/auth', json=body)

    def create_sensor(self, **kwargs):
        body = self.mk_attributes_body("sensor", None, kwargs)
        return self.post('sensor', json=body)

    def update_sensor(self, sensor_id, **kwargs):
        body = self.mk_attributes_body("sensor", sensor_id, kwargs)
        return self.patch('sensor/{}'.format(sensor_id), json=body)

    def delete_sensor(self, sensor_id):
        return self.delete('sensor/{}'.format(sensor_id))

    def get_sensors(self):
        return self.get("sensor")

    def get_sensor(self, sensor_id):
        return self.get('sensor/{}'.format(sensor_id))

    def get_sensor_timeseries(self, sensor_id, **kwargs):
        params = self.mk_timeseries_params(**kwargs)
        return self.get('sensor/{}/timeseries'.format(sensor_id),
                        params=params)

    def post_sensor_timeseries(self, sensor_id, **kwargs):
        body = self.mk_datapoint_body(**kwargs)
        return self.post('sensor/{}/timeseries'.format(sensor_id),
                         json=body)

    def live_sensor_timeseries(self, sensor_id, **kwargs):
        params = self.mk_timeseries_params(**kwargs)
        source = self.get('sensor/{}/timeseries/live'.format(sensor_id),
                          params=params, stream=True)
        return LiveService(source)

    def get_org(self):
        return self.get('organization')

    def update_org(self, **kwargs):
        body = self.mk_attributes_body("organization", None, kwargs)
        return self.patch('organization', json=body)

    def get_org_timeseries(self, **kwargs):
        params = self.mk_timeseries_params(**kwargs)
        return self.get('organization/timeseries', params=params)

    def live_org_timeseries(self, **kwargs):
        params = self.mk_timeseries_params(**kwargs)
        source = self.get('organization/timeseries/live', params=params)
        return LiveService(source)

    def post_org_timeseries(self, **kwargs):
        body = self.mk_datapoint_body(**kwargs)
        return self.post('organization/timeseries', json=body)

    def _get_json_path(self, json, path):
        try:
            return reduce(dict.__getitem__, path, json)
        except KeyError:
            return None

    def get_prev_page(self, json):
        prev_url = self._get_json_path(json, ["links", "prev"])
        return self.get(prev_url)

    def get_next_page(self, json):
        next_url = self._get_json_path(json, ["links", "next"])
        return self.get(next_url)

    def create_label(self, name=None):
        body = self.mk_attributes_body("label", None, {
            "name": name
        }) if name else None
        return self.post('label', json=body)

    def delete_label(self, label_id):
        return self.delete('label/{}'.format(label_id))

    def get_labels(self, **kwargs):
        params = self.mk_include_params(**kwargs)
        return self.get('label', params=params)

    def get_label(self, label_id, **kwargs):
        params = self.mk_include_params(**kwargs)
        return self.get('label/{}'.format(label_id), params=params)

    def get_label_sensors(self, label_id):
        return self.get('label/{}/relationships/sensor'.format(label_id))

    def update_label_sensors(self, label_id, sensor_ids):
        body = self.mk_relationships_body("sensor", sensor_ids)
        return self.patch('label/{}/relationships/sensor'.format(label_id),
                          json=body)

    def get_elements(self, **kwargs):
        params = self.mk_include_params(**kwargs)
        return self.get('element', params=params)

    def get_element(self, element_id, **kwargs):
        params = self.mk_include_params(**kwargs)
        return self.get('element/{}'.format(element_id), params=params)

    def update_element(self, element_id, **kwargs):
        body = self.mk_attributes_body("element", element_id, kwargs)
        return self.patch('element/{}'.format(element_id), json=body)

    def get_element_timeseries(self, element_id, **kwargs):
        params = self.mk_timeseries_params(**kwargs)
        return self.get('element/{}/timeseries'.format(element_id),
                        params=params)

    def post_element_timeseries(self, element_id, **kwargs):
        body = self.mk_datapoint_body(**kwargs)
        return self.post('element/{}/timeseries'.format(element_id),
                         json=body)

    def live_element_timeseries(self, element_id, **kwargs):
        params = self.mk_timeseries_params(**kwargs)
        source = self.get('element/{}/timeseries/live'.format(element_id),
                          params=params, stream=True)
        return LiveService(source)

    def _lua_uploads_from_files(self, files, **kwargs):
        def basename(f):
            return os.path.basename(f)

        def lua_file(name):
            return (basename(name), io.open(name, 'rb'), 'application/x-lua')

        main_file = kwargs.pop('main', None)
        # construct a dictionary of files that are not the main file
        files = {basename(name): lua_file(name) for name in files
                 if name != main_file}
        # then add the main file if given
        if main_file:
            files["user.lua"] = lua_file(main_file)
        return files

    def get_sensor_scripts(self):
        return self.get('sensor-script')

    def get_sensor_script(self, script_id):
        return self.get('sensor-script/{}'.format(script_id))

    def deploy_sensor_script(self, files, **kwargs):
        manifest = {
            "target": {
                "labels": kwargs.pop('label', []),
                "sensors": kwargs.pop('sensor', [])
            }
        }
        uploads = self._lua_uploads_from_files(files, **kwargs)
        uploads['manifest'] = ('manifest.json',
                               dump_json(manifest),
                               'application/json')
        return self.post('sensor-script', files=uploads)

    def get_cloud_scripts(self):
        return self.get('cloud-script')

    def get_cloud_script(self, script_id):
        return self.get('cloud-script/{}'.format(script_id))

    def get_cloud_script_timeseries(self, script_id, **kwargs):
        params = self.mk_timeseries_params(**kwargs)
        return self.get('cloud-script/{}/timeseries'.format(script_id),
                        params=params)

    def delete_cloud_script(self, script_id):
        return self.delete('cloud-script/{}'.format(script_id))

    def _mk_cloud_script_attributes(self, script_id, **kwargs):
        return self.mk_attributes_body("cloud-script", script_id, {
            "state": kwargs.pop("state", "running"),
            "name": kwargs.pop("name", None)
        })

    def update_cloud_script(self, script_id, **kwargs):
        body = self._mk_cloud_script_attributes(script_id, **kwargs)
        return self.patch('cloud-script/{}'.format(script_id), json=body)

    def deploy_cloud_script(self, files, **kwargs):
        uploads = self._lua_uploads_from_files(files, **kwargs)
        attributes = self._mk_cloud_script_attributes(None, **kwargs)
        uploads['attributes'] = ('attributes.json',
                                 dump_json(attributes),
                                 'application/json')
        return self.post('cloud-script', files=uploads)



import click
import helium
import dpath.util as dpath
import requests
from . import util


pass_service = click.make_pass_decorator(helium.Service)


@click.group()
def cli():
    """Operations on sensor-scripts.
    """
    pass


def _tabulate_scripts(result):
    def _map_sensor_count(json):
        targets = dpath.get(json, "relationships/status/data")
        return len(targets)

    def _map_progress(json):
        progress = dpath.values(json, "relationships/status/data/*/meta/progress")
        return sum(progress)/len(progress) if len(progress) > 0 else 0

    util.tabulate(result, [
        ('id', util.shorten_json_id),
        ('created', 'meta/created'),
        ('sensors', _map_sensor_count),
        ('state', 'meta/state'),
        ('progress', _map_progress),
        ('files', util.map_script_filenames)
    ])


def _tabulate_status(result):
    util.tabulate(result, [
        ('sensor', util.shorten_json_id),
        ('mac', 'meta/mac'),
        ('progress', 'meta/progress'),
        ('files', util.map_script_filenames)
    ])


@cli.command(name="list")
@pass_service
# renamed function to avoid collision with use of the built-in list function
def _list(service):
    """List all known sensor-scripts.
    """
    _tabulate_scripts(service.get_sensor_scripts().get('data'))


@cli.command()
@click.argument('script')
@pass_service
def status(service, script):
    """Get status for a script.

    Retrieves the current status for a sensor-script deploy request.
    """
    script = util.lookup_resource_id(service.get_sensor_scripts, script)
    script_info = service.get_sensor_script(script)
    _tabulate_status(dpath.get(script_info, 'data/relationships/status/data'))


@cli.command()
@click.argument('file', nargs=-1)
@click.option('--main', type=click.Path(exists=True),
              help="The main file for the script")
@click.option('-l', '--label', multiple=True,
              help="the id of a label")
@click.option('-s', '--sensor', multiple=True,
              help="the id of a sensor")
@click.option('-sf', '--sensor-file', type=click.File('rb'),
              help="the name of a file with sensor ids")
@pass_service
def deploy(service, file, sensor, sensor_file, label, main):
    """Deploy a sensor-script.

    Submit a deploy request of one or more FILEs. The targets for the
    deploy can be a combination of sensors or labels.

    The label (-l) and sensor (-s) specifier can be given multiple
    times to target multiple labels or sensors for a deploy.

    If the --main option is specified the given file is used as the
    `user.lua`, i.e. the main user script for the deployment. The file
    may be part of the list of files given to make it easier to
    specify wildcards.

    Note: One of the given files _may_ be called user.lua if the
    --main option is not given.  This file will be considered the
    primary script for the deploy.

    """
    sensor = list(sensor)
    if sensor_file:
        sensor.extend(sensor_file.read().splitlines())
    sensor = [util.lookup_resource_id(service.get_sensors, sensor_id)
              for sensor_id in sensor]
    deploy = service.deploy_sensor_script(file,
                                          label=label,
                                          sensor=sensor,
                                          main=main).get('data')
    _tabulate_scripts([deploy])


@cli.command()
@click.argument('script')
@click.argument('file')
@pass_service
def show(service, script, file):
    """Gets a script file from a given sensor-script.

    Fetches a FILE from a given sensor-SCRIPT.
    """
    script = util.lookup_resource_id(service.get_sensor_scripts, script)
    json = service.get_sensor_script(script).get('data')
    file_urls = [f.encode('utf-8') for f in dpath.get(json, 'meta/scripts')]
    names = dict(zip(util.extract_script_filenames(file_urls), file_urls))
    file_url = names[file]
    click.echo(requests.get(file_url).text)
